"""

    Module Models
    =============

    Text here
"""
import copy
from collections import OrderedDict
from copy import deepcopy
import pandas as pd
import json

import numpy as np
from solve import solvesas as solve

dtype = np.float64

from mesas.sas.specs import SAS_Spec
from collections import OrderedDict

import os

def _processinputs(input):
    if isinstance(input, dict):
        return input
    elif isinstance(input, str):
        if os.path.isfile(input):
            with open(input) as json_file:
                data = json.load(json_file) 
            return data
        else:
            raise FileNotFoundError(f'I could not find the input file: {input}')
    else:
        raise TypeError('input must be either a dict, or a path to a .json file')



class Model:
    '''
    Class for building and running StorAge Selection models

    An instance of this class must be constructed, populated with
    parameters (held in the dicts :ref:`sas_specs <sasspec>` and optionally :ref:`solute_parameters <solspec>`),
    associated with a dataset (`data_df`) and run using the `.run()` method.
    '''

    def __init__(self, data_df, config=None, sas_specs=None, solute_parameters=None, **kwargs):
        # defaults
        self._result = None
        self.jacobian = {}
        # load the timeseries data
        self.data_df = data_df
        # check for a configuration file
        if config:
            config = _processinputs(config)
        # process any options
        self._default_options = {
            'dt': 1,
            'verbose': False,
            'debug': False,
            'warning': True,
            'jacobian': False,
            'n_substeps': 1,
            'max_age': None,
            'sT_init': None,
            'influx': 'J',
            'ST_smallest_segment': 1./100,
            'ST_largest_segment': np.inf,
        }
        self._options = self._default_options
        components_to_learn = kwargs.get('components_to_learn')
        if config and 'options' in config.keys():
            self.options = config['options']
        self.options = kwargs
        # get the SAS specification
        if config and 'sas_specs' in config.keys():
            sas_specs = config['sas_specs']
        elif sas_specs:
            sas_specs = _processinputs(sas_specs)
        else:
            raise ValueError("No SAS specification found!")
        self.sas_specs = self.parse_sas_specs(sas_specs)
        self._numflux = len(self.sas_specs)
        self._fluxorder = list(self.sas_specs.keys())
        # get solute transport parameters
        if config and 'solute_parameters' in config.keys():
            solute_parameters = config['solute_parameters']
        elif solute_parameters:
            solute_parameters = _processinputs(solute_parameters)
        # defaults for solute transport
        self._default_parameters = {
            'mT_init': 0.,
            'C_old': 0.,
            'k1': 0.,
            'C_eq': 0.,
            'alpha': dict((flux, 1.) for flux in self._fluxorder),
            'observations': {}
        }
        self.solute_parameters = solute_parameters
        self.components_to_learn = components_to_learn

    def __repr__(self):
        """Creates a repr for the model"""
        repr = ''
        for flux, sas_spec in self.sas_specs.items():
            repr += f'flux = {flux}\n'
            repr += sas_spec.__repr__()
        return repr

    def parse_sas_specs(self, sas_specs):
        for flux, spec_in in sas_specs.items():
            spec = deepcopy(spec_in)
            assert flux in self.data_df.columns
            if isinstance(spec, SAS_Spec):
                sas_specs[flux] = spec
            else:
                assert isinstance(spec, dict)
                assert all([isinstance(component_spec, dict) for component_spec in spec.values()])
                sas_specs[flux] = SAS_Spec(spec, self.data_df)
        return sas_specs

    def copy_without_results(self):
        #returns a new instance of the model, but clears any results
        return Model(copy.deepcopy(self._data_df),
                     copy.deepcopy(self._sas_specs),
                     copy.deepcopy(self._solute_parameters),
                     copy.deepcopy(self._components_to_learn),
                     **copy.deepcopy(self._options))

    def subdivided_copy(self, flux, label, segment):
        """
        Creates a copy of the model with one segment of a sas function subdivided in two

        :param flux: name of the flux
        :param label: name of the component
        :param segment: segment (numbered from 0 for the youngest)
        :return: a new Model instance with one piecewise segment subdivided
        """

        # make a copy
        new_model = self.copy_without_results()

        # subdivide the component
        new_model.sas_specs[flux] = new_model.sas_specs[flux].subdivided_copy(label, segment)

        return new_model

    @property
    def fluxorder(self):
        return self._fluxorder

    @fluxorder.setter
    def fluxorder(self, new_fluxorder):
        raise AttributeError("fluxorder property is read-only")

    @property
    def solorder(self):
        return self._solorder

    @solorder.setter
    def solorder(self, new_solorder):
        raise AttributeError("solorder property is read-only")

    @property
    def result(self):
        """Results of running the sas model with the current parameters. See :ref:`results`
        """
        if self._result is None:
            raise AttributeError(
                "results are only defined once the model is run. Use .run() method to generate results ")
        else:
            return self._result

    @result.setter
    def result(self, result):
        raise AttributeError("Model results are read-only. Use .run() method to generate results ")

    @property
    def data_df(self):
        """Pandas dataframe holding the model inputs. See :ref:`inputs`"""
        return self._data_df

    @data_df.setter
    def data_df(self, new_data_df):
        if isinstance(new_data_df, pd.core.frame.DataFrame):
            self._data_df = new_data_df.copy()
        elif isinstance(new_data_df, str):
            self._data_df = pd.read_csv(new_data_df)
        else:
            raise TypeError('data_df must be either a pandas dataframe, or a path to a .csv file')
        self._timeseries_length = len(self._data_df)


    @property
    def options(self):
        """Options for running the model. See :ref:`options`
        """
        return self._options

    @options.setter
    def options(self, new_options):
        new_options = _processinputs(new_options)
        invalid_options = [optkey for optkey in new_options.keys()
                           if optkey not in self._default_options.keys()]
        if any(invalid_options):
            raise KeyError("Invalid options: {}".format(invalid_options))
        self._options.update(new_options)
        if self._options['max_age'] is None:
            self._options['max_age'] = self._timeseries_length
        assert self._options['max_age'] <= self._timeseries_length
        if self._options['sT_init'] is None:
            self._options['sT_init'] = np.zeros(self._options['max_age'])
        else:
            self._options['max_age'] = len(self._options['sT_init'])
        self._max_age = self._options['max_age']

    @property
    def sas_specs(self):
        """Specification of the SAS functions. See :ref:`sasspec`"""
        return self._sas_specs

    @sas_specs.setter
    def sas_specs(self, new_sas_specs):
        self._sas_specs = _processinputs(new_sas_specs)
        self._numflux = len(self._sas_specs)
        self._fluxorder = list(self._sas_specs.keys())

    def set_sas_spec(self, flux, sas_spec):
        self._sas_specs[flux] = sas_spec
        self._sas_specs[flux].make_spec_ts(self.data_df)

    def set_component(self, flux, component):
        label = component.label
        self._sas_specs[flux].components[label] = component
        self._sas_specs[flux].make_spec_ts()

    def set_sas_fun(self, flux, label, sas_fun):
        self._sas_specs[flux].components[label].sas_fun = sas_fun
        self._sas_specs[flux].make_spec_ts()

    def get_component_labels(self):
        component_labels = {}
        for flux in self._fluxorder:
            component_labels[flux] = list(self._sas_specs[flux].components.keys())
        return component_labels

    @property
    def components_to_learn(self):
        """
        A dictionary of lists giving the component labels that you want to train with MESAS

        The components in this dictionary can be set by assigning a dictionary. For example, this would
        include only the 'min' and 'max' components of the sas spec associated with the 'Q' flux.

            >>> model.components_to_learn = {'Q':['min', 'max']}

        The parameters associated with these components can be obtained as a 1-D array with:

            >>> seglist = model.get_parameter_list()

        and modified using:

            >>> model.update_from_parameter_list(seglst)

        :return:
        """
        return self._components_to_learn

    @components_to_learn.setter
    def components_to_learn(self, label_dict):
        if label_dict is not None:
            self._components_to_learn = OrderedDict((flux,
                                                     [label for label in self._sas_specs[flux]._componentorder if
                                                      label in label_dict[flux]]
                                                     ) for flux in self._fluxorder if flux in label_dict.keys())
            self._comp2learn_fluxorder = list(self._components_to_learn.keys())
            for flux in self._comp2learn_fluxorder:
                self._sas_specs[flux]._comp2learn_componentorder = self._components_to_learn[flux]

    def get_parameter_list(self):
        """
        Returns a concatenated list of segments from self.components_to_learn
        """
        return np.concatenate([self.sas_specs[flux].get_parameter_list() for flux in self._comp2learn_fluxorder])

    def update_from_parameter_list(self, parameter_list):
        """
        Modifies the components in self.components_to_learn from a concatenated list of segments
        """
        starti = 0
        for flux in self._comp2learn_fluxorder:
            nparams = len(self.sas_specs[flux].get_parameter_list())
            self.sas_specs[flux].update_from_parameter_list(parameter_list[starti: starti + nparams])
            starti += nparams

    @property
    def solute_parameters(self):
        """Parameters describing solute behavior. See :ref:`solspec`


        """
        return self._solute_parameters

    @solute_parameters.setter
    def solute_parameters(self, new_solute_parameters):
        # set parameters for solute transport
        # provide defaults if absent
        if new_solute_parameters is not None:
            new_solute_parameters = _processinputs(new_solute_parameters)
            self._solute_parameters = {}
            for sol, params in new_solute_parameters.items():
                self._solute_parameters[sol] = deepcopy(self._default_parameters)
                self.set_solute_parameters(sol, params)
            self._numsol = len(self._solute_parameters)
            self._solorder = list(self._solute_parameters.keys())
        else:
            self._numsol = 0
            self._solute_parameters = {}

    def set_solute_parameters(self, sol, params):
        invalid_parameters = [paramkey for paramkey in params.keys()
                              if paramkey not in self._default_parameters.keys()]
        if any(invalid_parameters):
            raise KeyError("invalid parameters for {}: {}".format(sol, invalid_parameters))
        self._solute_parameters[sol].update(params)

    def _create_solute_inputs(self):
        C_J = np.zeros((self._timeseries_length, self._numsol), dtype=dtype)
        mT_init = np.zeros((self._max_age, self._numsol), dtype=dtype)
        C_old = np.zeros(self._numsol, dtype=dtype)
        k1 = np.zeros((self._timeseries_length, self._numsol), dtype=dtype)
        C_eq = np.zeros((self._timeseries_length, self._numsol), dtype=dtype)
        alpha = np.ones((self._timeseries_length, self._numflux, self._numsol), dtype=dtype)
        if self.solute_parameters is not None:
            def _get_array(param, N):
                if param in self.data_df:
                    return self.data_df[param].values
                else:
                    return param * np.ones(N)

            for isol, sol in enumerate(self._solorder):
                C_J[:, isol] = self.data_df[sol].values
                C_old[isol] = self.solute_parameters[sol]['C_old']
                mT_init[:, isol] = _get_array(self.solute_parameters[sol]['mT_init'], self._max_age)
                k1[:, isol] = _get_array(self.solute_parameters[sol]['k1'], self._timeseries_length)
                C_eq[:, isol] = _get_array(self.solute_parameters[sol]['C_eq'],
                                           self._timeseries_length)
                for iflux, flux in enumerate(self._fluxorder):
                    alpha[:, iflux, isol] = _get_array(
                        self.solute_parameters[sol]['alpha'][flux], self._timeseries_length)
        return C_J, mT_init, C_old, alpha, k1, C_eq

    def _create_sas_lookup(self):
        nC_list = []
        nargs_list = []
        component_list = []
        component_type = []
        for flux in self._fluxorder:
            nC_list.append(len(self._sas_specs[flux].components))
            for component in self._sas_specs[flux]._componentorder:
                component_list.append(self.sas_specs[flux].components[component])
        nargs_list = [len(component.argsS) for component in component_list]
        component_type = [component.type for component in component_list]
        nC_total = np.sum(nC_list)
        nargs_total = np.sum(nargs_list)
        SAS_args = np.column_stack([[component.argsS] for component in component_list]).T
        P_list = np.column_stack([[component.argsP] for component in component_list]).T
        weights = np.column_stack([component.weights for component in component_list])
        return SAS_args, P_list, weights, component_type, nC_list, nC_total, nargs_list, nargs_total

    def run(self):
        """
        Call this method to run the model with current SAS specification, options, solute parameters, and timeseries dataframe. Results can then be accessed through the `.result` property of the model object
        :return: None
        """
        # water fluxes
        J = self.data_df[self.options['influx']].values
        Q = self.data_df[self._fluxorder].values
        sT_init = self.options['sT_init']
        timeseries_length = self._timeseries_length
        numflux = self._numflux
        #
        # SAS lookup table
        SAS_args, P_list, weights, component_type, nC_list, nC_total, nargs_list, nargs_total = self._create_sas_lookup()
        SAS_args = np.asfortranarray(SAS_args)
        P_list = np.asfortranarray(P_list)
        weights = np.asfortranarray(weights)
        #
        # Solutes
        C_J, mT_init, C_old, alpha, k1, C_eq = self._create_solute_inputs()
        numsol = self._numsol
        #
        # options
        dt = self.options['dt']
        verbose = self.options['verbose']
        debug = self.options['debug']
        warning = self.options['warning']
        jacobian = self.options['jacobian']
        n_substeps = self.options['n_substeps']
        max_age = self.options['max_age']

        # call the Fortran code
        fresult = solve(
            J, Q, SAS_args, P_list, weights,
            sT_init, dt, verbose, debug, warning, jacobian,
            mT_init, np.asfortranarray(C_J), np.asfortranarray(alpha), np.asfortranarray(k1),
            np.asfortranarray(C_eq), C_old,
            n_substeps, component_type, nC_list, nargs_list, numflux, numsol, max_age, timeseries_length, nC_total, nargs_total)
        sT, pQ, WaterBalance, mT, mQ, mR, C_Q, dsTdSj, dmTdSj, dCdSj, SoluteBalance = fresult

        if numsol > 0:
            self._result = {'C_Q': C_Q}
            for isol, sol in enumerate(self._solorder):
                for iflux, flux in enumerate(self._fluxorder):
                    colname = sol + ' --> ' + flux
                    self._data_df[colname] = C_Q[:, iflux, isol]
        else:
            self._result = {}
        self._result.update({'sT': np.moveaxis(sT, -1, 0), 'pQ': np.moveaxis(pQ, -1, 0),
                             'WaterBalance': np.moveaxis(WaterBalance, -1, 0), 'dsTdSj':np.moveaxis(dsTdSj, -1, 0)})
        if numsol > 0:
            self._result.update({'mT': np.moveaxis(mT, -1, 0), 'mQ': np.moveaxis(mQ, -1, 0), 'mR': np.moveaxis(mR, -1, 0),
            'SoluteBalance': np.moveaxis(SoluteBalance, -1, 0),
                                 'dmTdSj':np.moveaxis(dmTdSj, -1, 0), 'dCdSj':dCdSj})

    def get_jacobian(self, mode='segment', logtransform=True):
        J = None
        self.jacobian = {}
        for isol, sol in enumerate(self._solorder):
            if 'observations' in self.solute_parameters[sol]:
                self.jacobian[sol] = {}
                iP = 0
                for isolflux, solflux in enumerate(self._fluxorder):
                    if solflux in self.solute_parameters[sol]['observations']:
                        J_seg = None
                        for iflux, flux in enumerate(self._comp2learn_fluxorder):
                            for label in self._components_to_learn[flux]:
                                nP = len(self.sas_specs[flux].components[label].sas_fun.P)
                                J_S = np.squeeze(self.result['dCdSj'][:, iP:iP+nP, isolflux, isol])
                                if mode == 'endpoint':
                                    pass
                                elif mode == 'segment':
                                    # To get the derivative with respect to the segment length, we add up the derivative w.r.t. the
                                    # endpoints that would be displaced by varying that segment
                                    A = np.triu(np.ones(nP), k=0)
                                    J_S = np.dot(A, J_S.T).T
                                    if logtransform:
                                        J_S = J_S * self.sas_specs[flux].components[label].sas_fun._parameter_list
                                if J_seg is None:
                                    J_seg = J_S
                                else:
                                    J_seg = np.c_[J_seg, J_S]
                            PQ = np.sum(self.result['pQ'][:, :, iflux], axis=0)
                            J_old = 1 - PQ
                            J_old_sol = np.zeros((self._timeseries_length, self._numsol))
                            J_old_sol[:, list(self._solorder).index(sol)] = J_old.T
                            J_sol = np.c_[J_seg, J_old_sol]
                        if J is None:
                            J = J_sol
                        else:
                            J = np.concatenate((J, J_sol), axis=0)
                        self.jacobian[sol][flux] = {}
                        self.jacobian[sol][flux]['seg'] = J_seg
                        self.jacobian[sol][flux]['C_old'] = J_old
                    iP +=nP
        return J

    def get_residuals(self):
        ri = None
        for isol, sol in enumerate(self._solorder):
            if 'observations' in self.solute_parameters[sol]:
                for iflux, flux in enumerate(self._comp2learn_fluxorder):
                    if flux in self.solute_parameters[sol]['observations']:
                        obs = self.solute_parameters[sol]['observations'][flux]
                        C_obs = self.data_df[obs]
                        iflux = list(self._fluxorder).index(flux)
                        isol = list(self._solorder).index(sol)
                        this_ri = self.result['C_Q'][:, iflux, isol] - C_obs.values
                        if ri is None:
                            ri = this_ri
                        else:
                            ri = np.concatenate((ri, this_ri), axis=0)
                        self.data_df[f'residual {flux}, {sol}, {obs}'] = this_ri
        return ri

    def get_obs_index(self):
        index = None
        for isol, sol in enumerate(self._solorder):
            if 'observations' in self.solute_parameters[sol]:
                for iflux, flux in enumerate(self._comp2learn_fluxorder):
                    if flux in self.solute_parameters[sol]['observations']:
                        obs = self.solute_parameters[sol]['observations'][flux]
                        C_obs = self.data_df[obs]
                        this_index = ~np.isnan(C_obs.values)
                        if index is None:
                            index = this_index
                        else:
                            index = np.concatenate((index, this_index), axis=0)
        return np.nonzero(index)[0]



import numpy as np
np.seterr(divide='ignore', invalid='ignore')
import matplotlib.pyplot as plt
import matplotlib.patches as patches
import matplotlib.colors as colors
from collections import OrderedDict


def plot_transport_column(model, flux, sol, i=None, ax=None, dST=None, nST=20, cmap='cividis_r', TC_frac=0.3, vrange=None,
                          ST_max=None, omega_max=None, valvegap=0.3, hspace=0.015, artists_dict=OrderedDict(), do_init=True):
    dt = model.options['dt']
    Q = model.data_df[flux].iloc[i]
    C_old = model.solute_parameters[sol]['C_old']
    if i is None:
        i = 0

    sTs = np.r_[0, model.result['sT'][:-1, i]]
    mTs = np.r_[0, model.result['mT'][:-1, i, list(model._solorder).index(sol)]]
    sTe = model.result['sT'][:, i+1]
    mTe = model.result['mT'][:, i+1, list(model._solorder).index(sol)]
    sT = (sTs + sTe) / 2
    mT = (mTs + mTe) / 2
    pQ = model.result['pQ'][:, i, list(model._fluxorder).index(flux)]
    mQ = model.result['mQ'][:, i, list(model._fluxorder).index(flux), list(model._solorder).index(sol)]
    ST = np.r_[0, np.cumsum(sT)] * dt
    PQ = np.r_[0, np.cumsum(pQ)] * dt
    MT = np.r_[0, np.cumsum(mT)] * dt
    MQ = np.r_[0, np.cumsum(mQ)] * dt

    ST_mod = np.r_[0., model.sas_blends[flux].ST[i, :]]
    PQ_mod = np.r_[0., model.sas_blends[flux].P[i, :]]
    omega_mod = np.diff(PQ_mod) / np.diff(ST_mod)
    if omega_max is None:
        omega_max = np.nanmax(omega_mod) * 1.1
    if ST_max is None:
        ST_max = ST_mod[-1] * 1.1
    if np.isinf(ST_max):
        ST_max = ST.max() * 1.1

    CS = np.where(sT > 0, mT / sT, 0)

    if dST is None:
        ST_reg = np.linspace(0, ST_max, nST + 1)
    else:
        nST = int(ST_max / dST) + 1
        ST_reg = np.arange(nST + 1) * dST
    sT_reg = np.diff(ST_reg) / dt
    spacer = sT_reg[0] * dt * valvegap

    PQ_regmod = np.interp(ST_reg, ST_mod, PQ_mod, right=np.NaN)
    omega_reg = np.diff(PQ_regmod) / dt / sT_reg

    PQ_reg = np.interp(ST_reg, ST, PQ)
    pQ_reg = np.diff(PQ_reg)
    MQ_reg = np.interp(ST_reg, ST, MQ)
    any_old_reg = np.interp(ST_reg, ST, (PQ==PQ[-1]))[1:]
    f_old_reg = any_old_reg * np.diff(PQ_regmod - PQ_reg) / np.diff(PQ_regmod)

    CQ_reg_new = np.where(pQ_reg>0, np.diff(MQ_reg) / (Q * pQ_reg), 0)
    CQ_reg_new[np.isnan(CQ_reg_new)] = 0
    CQ_reg = CQ_reg_new * (1-f_old_reg) + C_old * f_old_reg


    cmap = plt.get_cmap(cmap)
    if vrange is None:
        vmin, vmax = min(CS.min(), C_old), max(CS.max(), C_old)
    else:
        vmin, vmax = vrange
    norm = colors.Normalize(vmin=vmin, vmax=vmax)

    if do_init:
        if ax is None:
            ax = plt.subplot(111)
        ax.xaxis.set_visible(False)
        ax.yaxis.set_visible(False)
        for loc, spine in ax.spines.items():
            spine.set_visible(False)
        ax.spines['right'].set_visible(False)
        ax1 = ax.inset_axes([0, 0, TC_frac, 1])
        ax2 = ax.inset_axes([TC_frac + hspace, 0, 1 - TC_frac - hspace, 1])
        ax1.set_visible(True)
        ax2.set_visible(True)

        ax1.spines['bottom'].set_visible(False)
        ax1.spines['top'].set_visible(False)
        ax1.spines['left'].set_position(('outward', 10))
        ax1.spines['right'].set_visible(False)
        ax1.xaxis.set_visible(False)

        ax2.spines['bottom'].set_position(('outward', 10))
        ax2.spines['top'].set_visible(False)
        ax2.spines['right'].set_visible(False)
        ax2.spines['left'].set_visible(False)
        ax2.yaxis.set_visible(False)

    if do_init:
        for j in range(len(sT)):
            artists_dict[f'TCpatch {j}'] = patches.Rectangle((0, 0), 0, 0, linewidth=0, edgecolor='None', visible=False)#, clip_on=False)
            ax1.add_patch(artists_dict[f'TCpatch {j}'])
        artists_dict[f'TCpatch C_old'] = patches.Rectangle((0, 0), 0, 0, linewidth=0, edgecolor='0.5', visible=False, hatch=r"///")#, clip_on=False)
        ax1.add_patch(artists_dict[f'TCpatch C_old'])

    for j in range(len(sT)):
        if sT[j] > 0:
            artists_dict[f'TCpatch {j}'].set_bounds((0, ST[j], 1, sT[j] * dt))
            artists_dict[f'TCpatch {j}'].set_facecolor(cmap(norm(CS[j])))
            artists_dict[f'TCpatch {j}'].set_visible(True)
    artists_dict[f'TCpatch C_old'].set_bounds((0, ST[-1], 1, ST_max - ST[-1]))
    artists_dict[f'TCpatch C_old'].set_facecolor(cmap(norm(C_old)))
    artists_dict[f'TCpatch C_old'].set_visible(True)

    if do_init:
        for n in range(nST):
            artists_dict[f'TQpatch {n}'] = patches.Rectangle((0, 0), 0, 0, linewidth=0, edgecolor='0.8', visible=False)#, clip_on=False)
            ax2.add_patch(artists_dict[f'TQpatch {n}'])

    for n in range(nST):
        artists_dict[f'TQpatch {n}'].set_bounds((0, ST_reg[n], omega_reg[n], sT_reg[n] * dt - spacer))
        artists_dict[f'TQpatch {n}'].set_facecolor(cmap(norm(CQ_reg[n])))
        artists_dict[f'TQpatch {n}'].set_visible(True)

    if do_init:
        omega_linestylestr = 'k--'
        artists_dict[f'omegaline v start'], = ax2.plot([0, 0], [0, 0], omega_linestylestr)#, clip_on=False)
        artists_dict[f'omegaline h start'], = ax2.plot([0, 0], [0, 0], omega_linestylestr)#, clip_on=False)
        for ip in range(len(omega_mod)):
            artists_dict[f'omegaline v {ip}'], = ax2.plot([0, 0], [0, 0], omega_linestylestr)#, clip_on=False)
            if ip > 0:
                artists_dict[f'omegaline h {ip}'], = ax2.plot([0, 0], [0, 0], omega_linestylestr)#, clip_on=False)
        artists_dict[f'omegaline h end'], = ax2.plot([0, 0], [0, 0], omega_linestylestr)#, clip_on=False)
        artists_dict[f'omegaline v end'], = ax2.plot([0, 0], [0, 0], omega_linestylestr)#, clip_on=False)

    artists_dict[f'omegaline v start'].set_data([0, 0], [0, ST_mod[0]])
    artists_dict[f'omegaline h start'].set_data([0, omega_mod[0]], [ST_mod[0], ST_mod[0]])
    for ip in range(len(omega_mod)):
        artists_dict[f'omegaline v {ip}'].set_data([omega_mod[ip], omega_mod[ip]], [ST_mod[ip], ST_mod[ip + 1]])
        if ip > 0:
            artists_dict[f'omegaline h {ip}'].set_data([omega_mod[ip - 1], omega_mod[ip]], [ST_mod[ip], ST_mod[ip]])
    artists_dict[f'omegaline h end'].set_data([omega_mod[-1], 0], [ST_mod[-1], ST_mod[-1]])
    artists_dict[f'omegaline v end'].set_data([0, 0], [ST_mod[-1], ST_max])

    if do_init:
        ax1.set_ylim([0, ST_max])
        ax2.set_ylim([0, ST_max])
        ax2.set_xlim([0, omega_max])
        ax1.invert_yaxis()
        ax2.invert_yaxis()
        ax1.set_ylabel('$S_T$')
        ax2.set_xlabel(f'$\omega(S_T)$ for {flux}')

        return ax1, ax2


def plot_influx(model, ax=None, sharex=None, i=None, artists_dict=OrderedDict(), do_init=True):
    J = model.data_df[model.options['influx']]
    if do_init:
        if ax is None:
            ax = plt.subplot(111, sharex=sharex)
        ax.plot(model.data_df.index, J, 'b')
        ax.set_ylim(bottom=0)
        artists_dict[f'plot_influx timeline'], = ax.plot(2*[model.data_df.index[0]], [0, 0], 'k', lw=0.5)
        ax.spines['right'].set_visible(False)
        ax.spines['top'].set_visible(False)
        ax.set_title(model.options['influx'])
    if i is not None:
        artists_dict[f'plot_influx timeline'].set_data(2*[model.data_df.index[i]], ax.get_ylim())


def plot_outflux(model, flux, ax=None, sharex=None, i=None, artists_dict=OrderedDict(), do_init=True):
    Q = model.data_df[flux]
    if do_init:
        if ax is None:
            ax = plt.subplot(111, sharex=sharex)
        ax.plot(model.data_df.index, Q, 'b')
        ax.set_ylim(bottom=0)
        artists_dict[f'plot_outflux timeline'], = ax.plot(2*[model.data_df.index[0]], [0, 0], 'k', lw=0.5)
        ax.spines['right'].set_visible(False)
        ax.spines['top'].set_visible(False)
        ax.set_title(flux)
    if i is not None:
        artists_dict[f'plot_outflux timeline'].set_data(2*[model.data_df.index[i]], ax.get_ylim())


def plot_influx_conc(model, sol, ax=None, sharex=None, i=None, artists_dict=OrderedDict(), do_init=True):
    C_J = model.data_df[sol]
    if do_init:
        if ax is None:
            ax = plt.subplot(111, sharex=sharex)
        ax.plot(model.data_df.index, C_J, 'b')
        artists_dict[f'plot_influx_conc timeline'], = ax.plot(2*[model.data_df.index[0]], [0, 0], 'k', lw=0.5)
        ax.spines['right'].set_visible(False)
        ax.spines['top'].set_visible(False)
        ax.set_title(sol)
    if i is not None:
        artists_dict[f'plot_influx_conc timeline'].set_data(2*[model.data_df.index[i]], ax.get_ylim())


def plot_outflux_conc(model, flux, sol, ax=None, sharex=None, i=None, artists_dict=OrderedDict(), do_init=True):
    colname = sol + ' --> ' + flux
    C_Q = model.data_df[colname]
    if do_init:
        if ax is None:
            ax = plt.subplot(111, sharex=sharex)
        ax.plot(model.data_df.index, C_Q, 'b')
        artists_dict[f'plot_outflux_conc timeline'], = ax.plot(2*[model.data_df.index[0]], [0, 0], 'k', lw=0.5)
        ax.spines['right'].set_visible(False)
        ax.spines['top'].set_visible(False)
        ax.set_title(colname)
    if i is not None:
        artists_dict[f'plot_outflux_conc timeline'].set_data(2*[model.data_df.index[i]], ax.get_ylim())


def plot_SAS_cumulative(model, flux, ax=None, sharex=None, i=None, artists_dict=OrderedDict(), do_init=True):
    if do_init:
        if ax is None:
            ax = plt.subplot(111, sharex=sharex)
        if i is None:
            i = 0
        artists_dict[f'plot_SAS {flux}'], = ax.plot(model.sas_blends[flux].ST[i, :], model.sas_blends[flux].P[i, :], 'bo-', lw=1.5)#, clip_on=False)
        ax.set_ylim([0, 1])
        ax.set_xlim(xmin=0)
        ax.plot(ax.get_xlim(), [1, 1], color='0.1', lw=0.8, ls=':')
        ax.plot(ax.get_xlim(), [0, 0], color='0.1', lw=0.8, ls=':')
        ax.spines['right'].set_visible(False)
        ax.spines['top'].set_visible(False)
        ax.set_xlabel('$S_T$')
        ax.set_ylabel('$\Omega(S_T)$')
        ax.spines['left'].set_position(('outward', 10))
        ax.spines['bottom'].set_position(('outward', 10))
    if i is not None:
        artists_dict[f'plot_SAS {flux}'].set_data(model.sas_blends[flux].ST[i, :], model.sas_blends[flux].P[i, :])


def plot_transport_column_with_timeseries(model, flux, sol, i=None, fig=None, artists_dict=OrderedDict(), **kwargs):

    if fig is None:
        fig = plt.figure(figsize=[11.5,  4.])
        fig.set_tight_layout(True)

    axTC = plt.subplot2grid((2, 3), (0,1), rowspan=2)
    axJ = plt.subplot2grid((2, 3), (0,0))
    axQ = plt.subplot2grid((2, 3), (0,2))
    axCJ = plt.subplot2grid((2, 3), (1,0))
    axCQ = plt.subplot2grid((2, 3), (1,2))

    plot_transport_column(model, flux, sol, i=i, ax=axTC, artists_dict=artists_dict, **kwargs)
    plot_influx(model, ax=axJ, i=i, artists_dict=artists_dict)
    plot_outflux(model, flux, ax=axQ, sharex=axJ, i=i, artists_dict=artists_dict)
    plot_influx_conc(model, sol, ax=axCJ, sharex=axJ, i=i, artists_dict=artists_dict)
    plot_outflux_conc(model, flux, sol, ax=axCQ, sharex=axJ, i=i, artists_dict=artists_dict)

    return axTC, axJ, axQ, axCJ, axCQ


def make_transport_column_animation(model, flux, sol, fig=None, frames=None, **kwargs):

    if frames is None:
        frames = range(model._timeseries_length-1)

    if fig is None:
        fig = plt.figure(figsize=[11.5,  4.])
        fig.set_tight_layout(True)

    from matplotlib.animation import FuncAnimation
    artists = OrderedDict()

    def init():
        i = 0
        plot_transport_column_with_timeseries(model, flux, sol, i=i, fig=fig, artists_dict=artists, **kwargs)
        return [artists[x] for x in artists.keys()]

    def update(frame):
        i=frame
        plot_transport_column_with_timeseries(model, flux, sol, i=i, fig=fig, artists_dict=artists, **kwargs)
        return [artists[x] for x in artists.keys()]

    return FuncAnimation(fig, update, frames=frames, init_func=init, blit=False)



"""

    ================
    Module functions
    ================

    This module defines classes representing SAS functions.

    Currently there are two classes : :class:`Piecewise` and :class:`Continuous`. The first allows the SAS function to be
    specified as a set of breakpoints in a piecewise linear form of the cumulative distribution. The second allows any
    distribution specified as a scipy.stats ``rv_continuous`` class to be used as a SAS function. However, it will be
    converted into a piecewise form when the model is run.

    Calling an instance of either class supplied with values of ST (as an array, list, or number) evaluates
    the CDF, and returns corresponding cumulative probabilities. The :func:`Piecewise.inv` method takes cumulative probabilities and
    returns values of ST.

"""

import copy

import matplotlib.pyplot as plt
import numpy as np
from scipy.interpolate import interp1d
from scipy.stats import rv_continuous
import scipy.stats
import importlib


typedict = {'gamma':1, 'beta':2}

class _SASFunctionBase:
    """
    Base function for SAS functions
    """

    _use = None

    def __init__(self):
        pass

    def _make_interpolators(self):
        """
        Create functions that can be called to return the CDF and inverse CDF

        Uses scipy.optimize.interp1d
        """
        self.interp1d_inv = interp1d(self.P, self.ST,
                                     fill_value=(self.ST_min, self.ST_max),
                                     kind='linear', copy=False,
                                     bounds_error=False, assume_sorted=True)
        self.interp1d = interp1d(self.ST, self.P,
                                 fill_value=(0., 1.),
                                 kind='linear', copy=False,
                                 bounds_error=False, assume_sorted=True)

    def __getitem__(self, i):
        return self

    def __setitem__(self, i):
        raise TypeError("You're not allowed to modify the SAS function for a specific index")

    # Next we define a number of properties. These act like attributes, but special functions
    # are called when the attributes are queried or assigned to

    #@property
    #def has_params(self):
    #    return self._has_params

    @property
    def argsS(self):
        return None

    @argsS.setter
    def argsS(self, new_args):
        raise NotImplementedError("Method for setting args directly has not been defined")
    @property

    def argsP(self):
        return None

    @argsP.setter
    def argsP(self, new_args):
        raise NotImplementedError("Method for setting args directly has not been defined")

    @property
    def ST(self):
        return self._ST

    @ST.setter
    def ST(self, new_ST):
        raise NotImplementedError("Method for setting ST directly has not been defined")

    @property
    def P(self):
        return self._P

    @P.setter
    def P(self, new_P):
        raise NotImplementedError("Method for setting P directly has not been defined")

    @property
    def parameter_list(self):
        return self._parameter_list

    @parameter_list.setter
    def parameter_list(self, new_parameter_list):
        raise NotImplementedError("Method for setting parameters directly has not been defined")

    def subdivided_copy(self, *args, **kwargs):
        raise TypeError("Cannot subdivide this type of function")

    def _convert_segment_list_to_ST(self, seglst):
        """converts a segment_list array into an array of ST endpoints"""
        return np.cumsum(seglst)

    def _convert_ST_to_segment_list(self, ST):
        """converts a segment_list array into an array of ST endpoints"""
        return np.r_[ST[0], np.diff(ST)]

    def inv(self, P):
        """
        The inverse Cumulative Distribution Function of the SAS function
        """
        raise NotImplementedError("Inverse CDF method not been defined")

    def __call__(self, ST):
        """
        The Cumulative Distribution Function of the SAS function
        """
        raise NotImplementedError("CDF method not been defined")

    def __repr__(self):
        """Return a repr of the SAS function"""
        raise NotImplementedError("__repr__ not been defined")

    def plot(self, ax=None, **kwargs):
        """Return a plot of the SAS function"""
        if ax is None:
            fig, ax = plt.subplots()
        return ax.plot(self.ST, self.P, **kwargs)

    def get_jacobian(self, dCdSj, index=None, mode='segment', logtransform=True):
        """ Calculates a limited jacobian of the model predictions """
        raise NotImplementedError("get_jacobian not been defined")


class Piecewise(_SASFunctionBase):
    """
    Class for piecewise-linear SAS functions

    You can define a piecewise sas function in several different ways. The default is a single segment from 0 to 1:

        >>>sas_fun = Piecewise()
        >>>sas_fun([0.1, 0.3, 0.9])
        array([0.1, 0.3, 0.9])

    The range of the single segment can be modified by providing ``ST_min`` and ``ST_max`` keywords

        >>>sas_fun = Piecewise(ST_min=100, ST_max=110)
        >>>sas_fun([99, 101, 103, 109, 111])
        array([0. , 0.1, 0.3, 0.9, 1. ])

    The (ST,P) pairs defining the CDF are given by the ``ST`` and ``P`` attributes

        >>>sas_fun.ST
        array([100.        , 103.00391164, 104.17022005,
        110.        ])
        >>>sas_fun.P
        array([0.        , 0.33333333, 0.66666667, 1.        ])

    The segments defining the function are stored in the ``segment_list``
    attribute as a numpy array:

        >>>seglst = sas_fun.segment_list
        array([100.        ,   3.00391164,   1.1663084 ,   5.82977995])

    The first item in the array is ``ST_min``, and each following value is the interval
    along the ST axis to the end of each segment.

    Once created, the values of ``segment_list`` or ``ST`` can be modified and the other will update too.
    Values of ``P`` can also be changed.

    """

    def __init__(self, ST=None, P=None, nsegment=1, ST_max=1., ST_min=0., auto='uniform'):
        """
        Initializes a Piecewise sas function.

        :param segment_list: Optional list of segment lengths in ST. First element is assumed to be ST_min.
        :param ST: Optional list of ST vales at the end of segments. Ignored if `segment_list` is passed
        :param P: Optional list of probabilities at the end of segments. Defaults to evenly spaced intervals
        :param nsegment: If neither `segment_list` nor `ST` are provided, the range between ST_min and ST_max is split into this many randomly, recursively split intervals.
        :param ST_max: The lower bound of the sas function support
        :param ST_min: The upper bound of the sas function support
        """

        self.ST_min = float(ST_min)
        self.ST_max = float(ST_max)
        #self._has_params = False

        # Note that the variables _ST, _P and _parameter_list are assigned to below
        # instead of their corresponding properties. This is to avoid triggering the _make_interpolators function
        # until the end

        if ST is not None:

            # Use the given ST values
            assert ST[0] >= 0
            assert np.all(np.diff(ST) > 0)
            assert len(ST) > 1
            self.nsegment = len(ST) - 1
            self._ST = np.array(ST, dtype=float)
            self._parameter_list = self._convert_ST_to_segment_list(self._ST)

        else:


            self.nsegment = nsegment

            if auto=='uniform':
                self._ST = np.linspace(self.ST_min, self.ST_max, self.nsegment + 1)
                self._parameter_list = self._convert_ST_to_segment_list(self._ST)
            elif auto=='random':
                # Generate a random function
                # Note this should probably be encapsulated in a function
                self._ST = np.zeros(self.nsegment + 1)
                ST_scaled = np.zeros(self.nsegment + 1)
                ST_scaled[-1] = 1.
                for i in range(self.nsegment - 1):
                    ST_scaled[self.nsegment - i - 1] = np.random.uniform(
                        0, ST_scaled[self.nsegment - i], 1)

                self._ST[:] = self.ST_min + (self.ST_max - self.ST_min) * ST_scaled
                self._parameter_list = self._convert_ST_to_segment_list(self._ST)

        # Make a list of probabilities P
        if P is not None:

            # Use the supplied values
            assert P[0] == 0
            assert P[-1] == 1
            assert np.all(np.diff(P) >= 0)
            assert len(P) == len(self._ST)
            self._P = np.r_[P]

        else:

            # Make P equally-spaced
            self._P = np.linspace(0, 1, self.nsegment + 1, endpoint=True)

        #self._has_params = True
        # call this to make the interpolation functions
        self._make_interpolators()


    @property
    def argsS(self):
        return self._ST

    @property
    def argsP(self):
        return self._P

    @_SASFunctionBase.ST.setter
    def ST(self, new_ST):
        try:
            assert new_ST[0] >= 0
            assert np.all(np.diff(new_ST) > 0)
            assert len(new_ST) > 1
        except Exception as ex:
            print('Problem with new ST')
            print(f'Attempting to set ST = {new_ST}')
            if not np.all(np.diff(new_ST) > 0):
                print(
                    "   -- if ST values are not distinct, try changing 'ST_largest_segment' and/or 'ST_smallest_segment'")
            raise ex
        self._ST = new_ST
        self.ST_min = self._ST[0]
        self.ST_max = self._ST[-1]
        self._parameter_list = self._convert_ST_to_segment_list(self._ST)
        self._make_interpolators()

    @_SASFunctionBase.P.setter
    def P(self, new_P):
        assert new_P[0] == 0
        assert new_P[-1] == 1
        assert len(new_P) == len(self._ST)
        self._P = new_P
        self._make_interpolators()

    @_SASFunctionBase.parameter_list.setter
    def parameter_list(self, new_parameter_list):
        self._parameter_list = new_parameter_list
        self.ST = self._convert_segment_list_to_ST(self._parameter_list)

    def subdivided_copy(self, segment, split_frac=0.5):
        """Returns a new Piecewise object instance with one segment divided into two

        :param segment: The segment to be split
        :param split_frac: (Optional) The fraction of the segment where the split is made. Default is 0.5
        """
        assert segment < self.nsegment
        P1 = self.P[segment]
        P2 = self.P[segment + 1]
        ST1 = self.ST[segment]
        ST2 = self.ST[segment + 1]
        P_new = P1 + split_frac * (P2 - P1)
        ST_new = ST1 + split_frac * (ST2 - ST1)
        P = np.insert(self.P, segment + 1, P_new)
        ST = np.insert(self.ST, segment + 1, ST_new)
        return Piecewise(ST=ST, P=P)

    def inv(self, P):
        """
        The inverse Cumulative Distribution Function of the SAS function

        :param P: Probabilities (can be a number, list or array-like)
        :return: A numpy array of corresponding ST values
        """
        P_arr = np.array(P)
        P_ravel = P_arr.ravel()
        return self.interp1d_inv(P_ravel).reshape(P_arr.shape)

    def __call__(self, ST):
        """
        The Cumulative Distribution Function of the SAS function

        :param ST: Age-ranked Storage values (can be a number, list or array-like)
        :return: A numpy array of corresponding probabilities
        """
        ST_arr = np.array(ST)
        ST_ravel = ST_arr.ravel()
        return self.interp1d(ST_ravel).reshape(ST_arr.shape)

    def __repr__(self):
        """Return a repr of the SAS function"""
        if not hasattr(self, 'nsegment'):
            return 'Not initialized'
        repr = '        ST: '
        for i in range(self.nsegment + 1):
            repr += '{ST:<10.4}  '.format(ST=self.ST[i])
        repr += '\n'
        repr += '        P : '
        for i in range(self.nsegment + 1):
            repr += '{P:<10.4}  '.format(P=self.P[i])
        repr += '\n'
        return repr

    def get_jacobian(self, dCdSj, index=None, mode='segment', logtransform=True):
        """
        Calculates a limited jacobian of the model predictions

        If :math:`\vec{y} = f(\vec{x})` the Jacobian is a matrix where each i, j entry is

        .. math:

            J_{i,j}=\frac{\partial y_i}{\partial x_j}

        ..

        Here, the :math:`y_i` values are the predicted concentrations at each timestep,
        and the :math:`x_j` are either the lengths of each piecewise segment along with ST_min (default),
        or the values of ST at the endpoints of each segment (with `mode='endpoint'`).

        The Jacobian is limited as it represents the sensitivity of the model predictions of output
        concentrations at each timestep to variations in the sas function at that timestep, and neglects
        the cumulative effects that changing the sas function would have on the state variables (ST and MS).

        :param dCdSj: jacobian produced by solve.f90
        :param index: index of timesteps to calculate the jacobian for
        :param mode: If 'endpoint' return the derivatives with respect to the ST of the endpoints,
        otherwise if 'segment' return the derivatives wrt the segment lengths
        :param logtransform: If True, return the derivatives with respect to the log transform of the segment lengths.
         Ignored if `mode='endpoint'. Default is True.
        :return: numpy array of the Jacobian

        """

        if index is None:
            index = np.arange(dCdSj.shape[0])

        J_S = dCdSj[index, :]

        if mode == 'endpoint':
            return J_S

        # To get the derivative with respect to the segment length, we add up the derivative w.r.t. the
        # endpoints that would be displaced by varying that segment
        A = np.triu(np.ones(self.nsegment + 1), k=0)
        J_seg = np.dot(A, J_S.T).T

        if logtransform:
            J_seglog = J_seg * self._parameter_list
            return J_seglog
        else:
            return J_seg


class Continuous(_SASFunctionBase):
    """
    Base function for SAS functions
    """

    _builtinfuncs = ['gamma', 'beta']
    def __init__(self, use, func, func_kwargs, P=None, nsegment=25, ST_max=1.797693134862315e+308):


        if use=='builtin' and func in self._builtinfuncs:
            self._func = getattr(scipy.stats, func)
        elif use=='scipy.stats' and isinstance(func, rv_continuous):
            self._func = func
        elif use=='scipy.stats' and isinstance(func, str):
            self._func = getattr(scipy.stats, func)
        else:
            raise Exception("'use' keyword must be either 'builtin' or 'scipy.stats'")
        self._use = use

        self._frozen_func = self._func(**func_kwargs)
        if self._use=='builtin':
            #self._has_params = True
            #self._ST = self.func.ppf(self._P)
            myargs = self._frozen_func.kwds.copy()
            if self._frozen_func.dist.name == 'gamma':
                self._argsS = []
                self._argsS += [myargs.pop('loc'), myargs.pop('scale')]
                self._argsS += [myargs.pop('a')]
            if self._frozen_func.dist.name == 'beta':
                self._argsS = []
                self._argsS += [myargs.pop('loc'), myargs.pop('scale')]
                self._argsS += [myargs.pop('a'), myargs.pop('b')]
            self._argsP = np.ones_like(self._argsS)*np.NaN
        elif self._use=='scipy.stats':
            # generate a piecewise approximation
            self.ST_max = np.float(ST_max)
            # Make a list of probabilities P
            if P is not None:
                # Use the supplied values
                self.nsegment = len(P) - 1
                self.P = np.r_[P]
            else:
                # Make P equally-spaced
                self.nsegment = nsegment
                self.P = np.linspace(0, 1, self.nsegment + 1, endpoint=True)
            self._argsS = self._ST
            self._argsP = self._P

    @property
    def argsS(self):
        return self._argsS

    @property
    def argsP(self):
        return self._argsP

    @property
    def func(self):
        return self._frozen_func

    @func.setter
    def func(selfs, new_func):
        raise ValueError("Function type cannot be changed")

    @_SASFunctionBase.ST.setter
    def ST(self, new_ST):
        try:
            assert new_ST[0] >= 0
            assert np.all(np.diff(new_ST) > 0)
            assert len(new_ST) > 1
        except Exception as ex:
            print('Problem with new ST')
            print(f'Attempting to set ST = {new_ST}')
            raise ex
        if new_ST[-1] > self.ST_max:
            self._ST = new_ST
            self.ST_max = self._ST[-1]
        else:
            self._ST = np.r_[new_ST, self.ST_max]
        self._P = self.func.cdf(self._ST)

    @_SASFunctionBase.P.setter
    def P(self, new_P):
        assert new_P[0] == 0
        assert new_P[-1] == 1
        assert np.all(np.diff(new_P) >= 0)
        self._P = new_P
        #if self._has_params:
        self._ST = self.func.ppf(self._P)

    def inv(self, P):
        """
        The inverse Cumulative Distribution Function of the SAS function
        """
        return self.func.ppf(P)

    def __call__(self, ST):
        """
        The Cumulative Distribution Function of the SAS function
        """
        return self.func.cdf(ST)

    def __repr__(self):
        """Return a repr of the SAS function"""
        repr = f'       {self._func.name} distribution\n'
        #if self._has_params:
        repr += f'        parameters: {self._frozen_func.args}\n'
        repr += '        Lookup table version:\n'
        repr += '          ST: '
        for i in range(self.nsegment + 1):
            repr += '{ST:<10.4}  '.format(ST=self.ST[i])
        repr += '\n'
        repr += '          P : '
        for i in range(self.nsegment + 1):
            repr += '{P:<10.4}  '.format(P=self.P[i])
        repr += '\n'
        #else:
            #repr = f'        (parameters not set)'
        return repr

    def get_jacobian(self, dCdSj, index=None, mode='segment', logtransform=True):
        """ Calculates a limited jacobian of the model predictions """
        raise NotImplementedError("get_jacobian not been defined")



import copy
from collections import OrderedDict

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from scipy.interpolate import interp1d
from collections.abc import Iterable

from mesas.sas.functions import Continuous, Piecewise, typedict

class SAS_Spec:

    def __init__(self, spec, data_df):
        """
        Initializes a sas spec
        """

        self.components = OrderedDict()
        is_only_component = len(spec) == 1
        for label, component_spec in spec.items():
            self.components[label] = Component(label, component_spec, data_df, is_only_component)
        self._componentorder = list(self.components.keys())
        self._comp2learn_componentorder = self._componentorder

    def make_spec_ts(self, data_df=None):
        """
        Create functions that can be called to return the CDF and inverse CDF
        """
        if data_df is None:
            data_df = self.model_data_df
        else:
            self.model_data_df = data_df

        # Trigger the method to create interpolators
        self.N = len(data_df)
        self._interp1d_inv = list(range(self.N))
        self._interp1d = list(range(self.N))
        self.ST_lists = list(range(self.N))
        self.P_lists = list(range(self.N))
        self.ST = None
        self.P = None
        for i in range(self.N):
            self.ST_lists[i] = np.sort(np.unique(np.concatenate(
                [component.sas_fun[i].ST for label, component in self.components.items()])))
            self.P_lists[i] = np.zeros_like(self.ST_lists[i])
            for label, component in self.components.items():
                self.P_lists[i] += component.sas_fun[i](self.ST_lists[i]) * component.weights.values[i]
            ST_min = self.ST_lists[i][0]
            ST_max = self.ST_lists[i][-1]
            self._interp1d_inv[i] = interp1d(self.P_lists[i], self.ST_lists[i],
                                             fill_value=(ST_min, ST_max),
                                             kind='linear', copy=False,
                                             bounds_error=False, assume_sorted=True)
            self._interp1d[i] = interp1d(self.ST_lists[i], self.P_lists[i],
                                         fill_value=(0., 1.),
                                         kind='linear', copy=False,
                                         bounds_error=False, assume_sorted=True)
        # create matricies of ST, P padded with extra values on the left
        maxj = np.max([len(ST) for ST in self.ST_lists])
        self.ST = np.zeros((maxj, self.N))
        self.P = np.zeros_like(self.ST)
        for i in range(self.N):
            Nj = len(self.ST_lists[i])
            self.ST[maxj - Nj:maxj, i] = self.ST_lists[i]
            self.P[maxj - Nj:maxj, i] = self.P_lists[i]
            if Nj < maxj:
                self.ST[:maxj - Nj, i] = -1 - np.arange(maxj - Nj)
                self.P[:maxj - Nj, i] = np.zeros(maxj - Nj)
        self.ST = self.ST.T
        self.P = self.P.T

    def __call__(self, ST, i):
        """
        The Cumulative Distribution Function of the SAS function at a timestep

        :param ST: Age-ranked Storage values (can be a number, list or array-like)
        :param i: index of the timestep
        :return: A numpy array of corresponding probabilities
        """
        return self._interp1d[i](ST)

    def inv(self, P, i):
        """
        The inverse Cumulative Distribution Function of the SAS function at a timestep

        :param P: Probabilities (can be a number, list or array-like)
        :param i: index of the timestep
        :return: A numpy array of corresponding ST values
        """
        return self._interp1d_inv[i](P)

    def __repr__(self):
        """produce a string representation of the spec"""
        repr = ''
        for label, component in self.components.items():
            repr += component.__repr__()
        return repr

    def get_parameter_list(self):
        """
        Returns a list of the parameters in each component sas function
        """
        return np.concatenate(
            [self.components[label].sas_fun.parameter_list for label in self._comp2learn_componentorder])

    def update_from_parameter_list(self, parameter_list):
        """Modify the component sas functions using a modified parameter_list"""
        starti = 0
        for label in self._comp2learn_componentorder:
            component = self.components[label]
            nparams = len(component.sas_fun.parameter_list)
            component.sas_fun.parameter_list = parameter_list[starti: starti + nparams]
            starti += nparams
        self.make_spec_ts()

    def plot(self, ax=None, **kwargs):
        """Plot the component sas_functions"""
        if ax is None:
            fig, ax = plt.subplots()
        componentlines = dict([(label, component.plot(ax=ax, **kwargs))
                               for label, component in self.components.items()])
        ax.set_xlabel('$S_T$')
        ax.set_ylabel('$\Omega(S_T)$')
        ax.legend(frameon=False)
        return componentlines

    def get_jacobian(self, *args, index=None, **kwargs):
        """
        Get the jacobian of each component sas function

        Calls the get_jacobian method of each component sas function.
        Components are concatenated columnwise and returned as a single numpy array.
        """
        if index is None:
            index = np.arange(self.N)
        cat_me = [self.components[label].sas_fun.get_jacobian(*args, index=index, **kwargs).T *
                  self.components[label].weights.values[index]
                  for label in self._comp2learn_componentorder]
        return np.concatenate(cat_me, axis=0).T


class Component:
    """
    The Component class defines simple objects that package together sas
    functions with their weights timeseries and a label. The class also defines __repr__ and plot functions
    that call the corresponding functions of the underlying sas functions.
    """

    def __init__(self, label, spec, data_df, is_only_component):
        self.label = label
        self._spec = spec.copy()
        self._data_df = data_df
        self.N = len(data_df)
        self.weights = pd.Series(index=data_df.index, data=1.) if is_only_component else data_df[label]
        ST = self.expand(spec.pop('ST')) if 'ST' in spec else _NoneList()
        P = self.expand(spec.pop('P')) if 'P' in spec else _NoneList()
        if 'args' in spec:
            assert isinstance(spec['args'], dict)
            argdict = OrderedDict()
            for arg, value in spec.pop('args').items():
                argdict[arg] = self.expand(value)
            self._args = [dict(zip(argdict, values)) for values in zip(*argdict.values())]
            if 'scipy.stats' in spec:
                use = 'scipy.stats'
                func = spec.pop('scipy.stats')
            else:
                use = spec.pop('use', 'builtin')
                func = spec.pop('func')
            if use == 'scipy.stats':
                self.type = -1
            else:
                assert use =='builtin'
                self.type = typedict[func]
            self._sas_funs = [Continuous(use, func, func_kwargs=self._args[i], P=P[i, :], **spec)
                                  for i in range(self.N)]
        elif ST is not None:
            self.type = -1
            self._sas_funs = [Piecewise(ST=ST[i, :], P=P[i, :], **spec) for i in range(self.N)]
        else:
            raise Exception(f"Insufficient information to specify SAS function for {label}")


    def expand(self, value):
        if isinstance(value, Iterable) and not isinstance(value, str):
            return np.concatenate([[self.expand(x)] for x in value], axis=0).T
        else:
            if isinstance(value, str):
                return self._data_df[value].values
            else:
                return float(value) * np.ones(self.N)

    @property
    def ST(self):
        return np.concatenate([[sas_fun.ST] for sas_fun in self._sas_funs]).T

    @property
    def P(self):
        return np.concatenate([[sas_fun.P] for sas_fun in self._sas_funs]).T

    @property
    def argsS(self):
        return np.concatenate([[sas_fun.argsS] for sas_fun in self._sas_funs]).T

    @property
    def argsP(self):
        return np.concatenate([[sas_fun.argsP] for sas_fun in self._sas_funs]).T

    @property
    def sas_fun(self):
        return [sas_fun for sas_fun in self._sas_funs]

    def __getitem__(self, i):
        return self._sas_funs[i]

    def __repr__(self):
        repr = ''
        repr += self._spec.__repr__()
        #repr += '    component = {}'.format(self.label) + '\n'
        #for i in range(self.N):
            #repr += self[i].__repr__()
        return repr

    def plot(self, *args, **kwargs):
        return self.sas_fun.plot(*args, label=self.label, **kwargs)


class _NoneList:
    def __init__(self):
        pass
    def __getitem__(self, *args, **kwargs):
        return None



